
/*

cellwriter -- a character recognition input method
Copyright (C) 2007 Michael Levin <risujin@risujin.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include <gtk/gtk.h>
#include <math.h>

/*
        Unicode blocks
*/

typedef struct {
        short enabled;
        const int start, end;
        const char *name;
} UnicodeBlock;

extern UnicodeBlock unicode_blocks[];

/*
        Profile
*/

extern int profile_line;

const char *profile_read(void);
int profile_write(const char *str);
int profile_sync_int(int *var);
int profile_sync_short(short *var);

/*
        GTK/GDK/Glib specific
*/

/* Multiply to convert RGB to GDK color */
#define COLOR_SCALE 256

/* Constants may not have been defined if GLib is not included */
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#ifndef NULL
#define NULL ((void*)0)
#endif

/* A macro used to initialize GdkColor with RGB values */
#define RGB_TO_GDKCOLOR(r, g, b) {0, (r) * 256, (g) * 256, (b) * 256 }

static inline void cairo_set_source_gdk_color(cairo_t *cairo,
                                              const GdkColor *color,
                                              double alpha)
/* Set the cairo source color from a GdkColor */
{
        cairo_set_source_rgba(cairo, color->red / 65535.,
                                     color->green / 65535.,
                                     color->blue / 65535., alpha);
}

static inline void cairo_pattern_add_gdk_color_stop(cairo_pattern_t *pattern,
                                                    double offset,
                                                    GdkColor *color,
                                                    double alpha)
/* Add a GdkColor color stop to a cairo pattern */
{
        cairo_pattern_add_color_stop_rgba(pattern, offset,
                                          color->red / 65535.,
                                          color->green / 65535.,
                                          color->blue / 65535., alpha);
}

void highlight_gdk_color(const GdkColor *base, GdkColor *out, int value);

/*
        Variable argument parsing
*/

#ifdef _EFISTDARG_H_
char *nvav(int *plen, const char *format, va_list va);
#endif
char *nva(int *length, const char *format, ...);
char *va(const char *format, ...);

/*
        Angles
*/

/* Size of the ANGLE data type in bytes */
#define ANGLE_SIZE 2

#if (ANGLE_SIZE == 4)

/* High-precision angle type */
typedef int ANGLE;
#define ANGLE_PI 2147483648

#elif (ANGLE_SIZE == 2)

/* Medium-precision angle type */
typedef short ANGLE;
#define ANGLE_PI 32768

#else

/* Low-precision angle type */
typedef signed char ANGLE;
#define ANGLE_PI 128

#endif

/*
        2D Vector
*/

typedef struct Vec2 {
	float x, y;
} Vec2;

static inline void vec2_set(Vec2 *dest, float x, float y)
{
	dest->x = x;
	dest->y = y;
}
#define vec2_from_coords vec2_set

static inline void vec2_copy(Vec2 *dest, const Vec2 *src)
{
	dest->x = src->x;
	dest->y = src->y;
}

static inline void vec2_sub(Vec2 *dest, const Vec2 *a, const Vec2 *b)
{
	dest->x = a->x - b->x;
	dest->y = a->y - b->y;
}

static inline void vec2_sum(Vec2 *dest, const Vec2 *a, const Vec2 *b)
{
	dest->x = a->x + b->x;
	dest->y = a->y + b->y;
}

static inline float vec2_dot(const Vec2 *a, const Vec2 *b)
{
	return a->x * b->x + a->y * b->y;
}

static inline float vec2_cross(const Vec2 *a, const Vec2 *b)
{
	return a->y * b->x - b->y * a->x;
}

static inline void vec2_scale(Vec2 *dest, const Vec2 *src, float scale)
{
	dest->x = src->x * scale;
	dest->y = src->y * scale;
}

static inline void vec2_avg(Vec2 *dest, const Vec2 *a, const Vec2 *b,
			    float scale)
{
	dest->x = a->x + (b->x - a->x) * scale;
	dest->y = a->y + (b->y - a->y) * scale;
}

static inline float vec2_square(const Vec2 *src)
{
        return src->x * src->x + src->y * src->y;
}

#ifdef _MATH_H
static inline float vec2_mag(const Vec2 *src)
{
	return sqrt(src->x * src->x + src->y * src->y);
}

static inline ANGLE vec2_angle(const Vec2 *src)
{
	return (ANGLE)(atan2f(src->y, src->x) * ANGLE_PI / M_PI + 0.5f);
}

static inline float vec2_norm(Vec2 *dest, const Vec2 *a)
{
	float mag = vec2_mag(a);
	dest->x = a->x / mag;
	dest->y = a->y / mag;
	return mag;
}

static inline void vec2_proj(Vec2 *dest, const Vec2 *a, const Vec2 *b)
{
	float dist = vec2_dot(a, b), mag = vec2_mag(b), mag2 = mag * mag;
	dest->x = dist * b->x / mag2;
	dest->y = dist * b->y / mag2;
}

static inline void vec2_from_angle(Vec2 *dest, ANGLE angle, float mag)
{
	dest->y = sinf(angle * M_PI / ANGLE_PI) * mag;
	dest->x = cosf(angle * M_PI / ANGLE_PI) * mag;
}
#endif

